@@ -1,4 +1,5 @@
 
+#include "include/stringify.h"
 #include "CrushTester.h"
 
 #include <algorithm>
@@ -355,14 +356,17 @@ void CrushTester::write_integer_indexed_scalar_data_string(vector<string> &dst,
   dst.push_back( data_buffer.str() );
 }
 
-int CrushTester::test_with_crushtool()
+int CrushTester::test_with_crushtool(const string& crushtool,
+                                     int timeout)
 {
+  string timeout_string = stringify(timeout);
   vector<const char *> cmd_args;
-  cmd_args.push_back("crushtool");
+  cmd_args.push_back("timeout");
+  cmd_args.push_back(timeout_string.c_str());
+  cmd_args.push_back(crushtool.c_str());
   cmd_args.push_back("-i");
   cmd_args.push_back("-");
   cmd_args.push_back("--test");
-  cmd_args.push_back("--output-csv");
   cmd_args.push_back(NULL);
 
   int pipefds[2];
@@ -413,9 +417,13 @@ int CrushTester::test_with_crushtool()
     // major success!
     return 0;
   }
+  if (r == 124) {
+    // the test takes longer than timeout and was interrupted
+    return -EINTR;
+  }
 
   if (r == ENOENT) {
-    err << "unable to find 'crushtool' to test the map";
+    err << "unable to find " << cmd_args << " to test the map";
     return -ENOENT;
   }
 
