@@ -1,8 +1,17 @@
+// -*- mode:C++; tab-width:8; c-basic-offset:2; indent-tabs-mode:t -*-
+// vim: ts=8 sw=2 smarttab
+
 #include <sys/types.h>
 #include <netinet/in.h>
 #include <arpa/nameser.h>
 #include <resolv.h>
 
+#include "acconfig.h"
+
+#ifdef HAVE_ARPA_NAMESER_COMPAT_H
+#include <arpa/nameser_compat.h>
+#endif
+
 #include "rgw_common.h"
 #include "rgw_resolve.h"
 
@@ -161,7 +170,7 @@ RGWResolver::RGWResolver() {
 
 int RGWResolver::resolve_cname(const string& hostname, string& cname, bool *found) {
   return resolver->resolve_cname(hostname, cname, found);
-};
+}
 
 RGWResolver *rgw_resolver;
 
@@ -170,3 +179,8 @@ void rgw_init_resolver()
 {
   rgw_resolver = new RGWResolver();
 }
+
+void rgw_shutdown_resolver()
+{
+  delete rgw_resolver;
+}
