@@ -1,3 +1,6 @@
+// -*- mode:C++; tab-width:8; c-basic-offset:2; indent-tabs-mode:t -*-
+// vim: ts=8 sw=2 smarttab
+
 #ifndef CEPH_RGW_REST_H
 #define CEPH_RGW_REST_H
 #define TIME_BUF_SIZE 128
@@ -10,10 +13,10 @@
 
 extern std::map<std::string, std::string> rgw_to_http_attrs;
 
-extern void rgw_rest_init(CephContext *cct);
-
 extern string lowercase_dash_http_attr(const string& orig);
 
+extern void rgw_rest_init(CephContext *cct, RGWRegion& region);
+
 extern void rgw_flush_formatter_and_reset(struct req_state *s,
 					 ceph::Formatter *formatter);
 
@@ -354,19 +357,14 @@ public:
   }
 };
 
-static const int64_t NO_CONTENT_LENGTH = -1;
-
 extern void set_req_state_err(struct req_state *s, int err_no);
 extern void dump_errno(struct req_state *s);
 extern void dump_errno(struct req_state *s, int ret);
-extern void end_header(struct req_state *s,
-                       RGWOp *op = NULL,
-                       const char *content_type = NULL,
-                       const int64_t proposed_content_length = NO_CONTENT_LENGTH,
-		       bool force_content_type = false);
+extern void end_header(struct req_state *s, RGWOp *op = NULL, const char *content_type = NULL);
 extern void dump_start(struct req_state *s);
 extern void list_all_buckets_start(struct req_state *s);
 extern void dump_owner(struct req_state *s, string& id, string& name, const char *section = NULL);
+extern void dump_string_header(struct req_state *s, const char *name, const char *val);
 extern void dump_content_length(struct req_state *s, uint64_t len);
 extern void dump_etag(struct req_state *s, const char *etag);
 extern void dump_epoch_header(struct req_state *s, const char *name, time_t t);
